/*
	taddump_tiny.c

	original code: SASANO Takayoshi
	--- public domain, no warranty.
*/
#include <btron/btron.h>
#include <errcode.h>
#include <bstdio.h>
#include <bstdlib.h>

typedef	struct {
	UB	*str;
	W	id;
} SegName;

LOCAL	const SegName	NameTbl[] = {
	{"TS_INFO",	0xE0},
	{"TS_TEXT",	0xE1},
	{"TS_TEXTEND",	0xE2},
	{"TS_FIG",	0xE3},
	{"TS_FIGEND",	0xE4},
	{"TS_IMAGE",	0xE5},
	{"TS_VOBJ",	0xE6},
	{"TS_DFUSEN",	0xE7},
	{"TS_FFUSEN",	0xE8},
	{"TS_SFUSEN",	0xE9},
	{"TS_TPAGE",	0xA0},
	{"TS_TRULER",	0xA1},
	{"TS_TFONT",	0xA2},
	{"TS_TCHAR",	0xA3},
	{"TS_TATTR",	0xA4},
	{"TS_TSTYLE",	0xA5},
	{"TS_TVAR",	0xAD},
	{"TS_TMEMO",	0xAE},
	{"TS_TAPPL",	0xAF},
	{"TS_FPRIM",	0xB0},
	{"TS_FDEF",	0xB1},
	{"TS_FGRP",	0xB2},
	{"TS_FMAC",	0xB3},
	{"TS_FATTR",	0xB4},
	{"TS_FPAGE",	0xB5},
	{"TS_FMEMO",	0xBE},
	{"TS_FAPPL",	0xBF},
	{"unknown",	~0},
};

/* find segment name (slow...) */
LOCAL	UB	*find_segname(W id)
{
	W	i;

	for (i = 0; i < sizeof(NameTbl) / sizeof(SegName) - 1; i++) {
		if (id == NameTbl[i].id) break;
	}

	return NameTbl[i].str;
}

/* dump */
LOCAL	void	dump(UH *buf, W size)
{
	W	i, pos, id, len;

	for (i = 0; i < size; i++) {
		if ((id = buf[(pos = i)] ) < 0xff00) continue;

		if (buf[pos + 1] == 0xffff) {
			len = buf[pos + 2] | (buf[pos + 3] << 16);
			i += 3;
		} else {
			len = buf[pos + 1];
			i += 1;
		}

		printf("%08x: %04x %08x <%s>\n",
		       pos * 2, id, len, find_segname(id & 0xff));

		i += len / 2;
	}
}


/* main */
EXPORT	W	main(W ac, TC *av[])
{
	W	err, fd, size;
	LINK	l;
	VP	buf;

	/* $B0z?t$N%A%'%C%/(B */
	if (ac < 2) {
		printf("usage: %S [filename]\n", av[0]);
		err = ER_PAR;
		goto fin0;
	}

	/* $BF~NO%U%!%$%k$X$N%j%s%/<h$j=P$7(B */
	err = get_lnk(av[1], &l, F_NORM);
	if (err < ER_OK) {
		printf("main: get_lnk %d\n", err);
		goto fin0;
	}

	/* $BF~NO%U%!%$%k$N%*!<%W%s(B */
	err = opn_fil(&l, F_READ | F_EXCL, NULL);
	if (err < ER_OK) {
		printf("main: opn_fil %d\n", err);
		goto fin0;
	}
	fd = err;

	/* $B<g(BTAD$B%l%3!<%I$X0\F0(B */
	err = fnd_rec(fd, F_ENDTOP, (1 << 1), 0, NULL);
	if (err < ER_OK) {
		printf("main: fnd_rec %d\n", err);
		goto fin1;
	}

	/* $B%l%3!<%I%5%$%:$N<hF@(B */
	err = rea_rec(fd, 0, NULL, 0, &size, NULL);
	if (err < ER_OK) {
		printf("main: rea_rec %d\n", err);
		goto fin1;
	}

	/* $B%a%b%j$N3NJ](B */
	buf = malloc(size);
	if (buf == NULL) {
		printf("main: malloc NULL\n");
		err = ER_NOMEM;
		goto fin1;
	}

	/* $B%l%3!<%IA4BN$NFI$_9~$_(B */
	err = rea_rec(fd, 0, buf, size, NULL, NULL);
	if (err < ER_OK) {
		printf("main: rea_rec %d\n", err);
		goto fin2;
	}

	/* $B%@%s%W$r3+;O(B */
	dump(buf, size / 2);

 fin2:
	free(buf);
 fin1:
	cls_fil(fd);
 fin0:
	return err;
}
